/*

 Header for PLY polygon files.

  - Greg Turk, March 1994

   A PLY file contains a single polygonal _object_.

    An object is composed of lists of _elements_.  Typical elements are
    vertices, faces, edges and materials.

     Each type of element for a given object has one or more _properties_
     associated with the element type.  For instance, a vertex element may
     have as properties three floating-point values x,y,z and three unsigned
     chars for red, green and blue.

      ---------------------------------------------------------------

       Copyright (c) 1994 The Board of Trustees of The Leland Stanford
       Junior University.  All rights reserved.

        Permission to use, copy, modify and distribute this software and its
        documentation for any purpose is hereby granted without fee, provided
        that the above copyright notice and this permission notice appear in
        all copies of this software and that you do not sell the software.

         THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,
         EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
         WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.

*/

#ifndef PLY_FILE_INCLUDED
#define PLY_FILE_INCLUDED


/*#ifndef WIN32
#ifdef __cplusplus
extern "C" {
#endif
#endif //WIN32*/

#include "ply_struct.h"

#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <string>

#define PLY_ASCII         1      /* ascii PLY file */
#define PLY_BINARY_BE     2      /* binary PLY file, big endian */
#define PLY_BINARY_LE     3      /* binary PLY file, little endian */
#define PLY_BINARY_NATIVE 4      /* binary PLY file, same endianness as current architecture */

#define PLY_OKAY    0           /* ply routine worked okay */
#define PLY_ERROR  -1           /* error in ply routine */

    /* scalar data types supported by PLY format */

#define PLY_START_TYPE 0
#define PLY_CHAR       1
#define PLY_SHORT      2
#define PLY_INT        3
#define PLY_UCHAR      4
#define PLY_USHORT     5
#define PLY_UINT       6
#define PLY_FLOAT      7
#define PLY_DOUBLE     8
#define PLY_INT_8      9
#define PLY_UINT_8     10
#define PLY_INT_16     11
#define PLY_UINT_16    12
#define PLY_INT_32     13
#define PLY_UINT_32    14
#define PLY_FLOAT_32   15
#define PLY_FLOAT_64   16

#define PLY_END_TYPE   17

#define  PLY_SCALAR  0
#define  PLY_LIST    1

#define PLY_STRIP_COMMENT_HEADER 0

typedef struct PlyProperty {    /* description of a property */

    std::string name;                     /* property name */
    int external_type;                    /* file's data type */
    int internal_type;                    /* program's data type */
    int offset;                           /* offset bytes of prop in a struct */

    int is_list;                          /* 1 = list, 0 = scalar */
    int count_external;                   /* file's count type */
    int count_internal;                   /* program's count type */
    int count_offset;                     /* offset byte for list count */

} PlyProperty;

typedef struct PlyElement {     /* description of an element */
    char *name;                   /* element name */
    int num;                      /* number of elements in this object */
    int size;                     /* size of element (bytes) or -1 if variable */
    int nprops;                   /* number of properties for this element */
    PlyProperty **props;          /* list of properties in the file */
    char *store_prop;             /* flags: property wanted by user? */
    int other_offset;             /* offset to un-asked-for props, or -1 if none*/
    int other_size;               /* size of other_props structure */
} PlyElement;

typedef struct PlyOtherProp {   /* describes other properties in an element */
    char *name;                   /* element name */
    int size;                     /* size of other_props */
    int nprops;                   /* number of properties in other_props */
    PlyProperty **props;          /* list of properties in other_props */
} PlyOtherProp;

typedef struct OtherData { /* for storing other_props for an other element */
    void *other_props;
} OtherData;

typedef struct OtherElem {     /* data for one "other" element */
    char *elem_name;             /* names of other elements */
    int elem_count;              /* count of instances of each element */
    OtherData **other_data;      /* actual property data for the elements */
    PlyOtherProp *other_props;   /* description of the property data */
} OtherElem;

typedef struct PlyOtherElems {  /* "other" elements, not interpreted by user */
    int num_elems;                /* number of other elements */
    OtherElem *other_list;        /* list of data for other elements */
} PlyOtherElems;

typedef struct PlyFile {        /* description of PLY file */
    FILE *fp;                     /* file pointer */
    int file_type;                /* ascii or binary */
    float version;                /* version number of file */
    int nelems;                   /* number of elements of object */
    PlyElement **elems;           /* list of elements */
    int num_comments;             /* number of comments */
    char **comments;              /* list of comments */
    int num_obj_info;             /* number of items of object information */
    char **obj_info;              /* list of object info items */
    PlyElement *which_elem;       /* which element we're currently writing */
    PlyOtherElems *other_elems;   /* "other" elements from a PLY file */
} PlyFile;

    /* memory allocation */
extern char *my_alloc();
#define myalloc(mem_size) my_alloc((mem_size), __LINE__, __FILE__)

#ifndef ALLOCN
#define REALLOCN(PTR,TYPE,OLD_N,NEW_N)							\
{										\
    if ((OLD_N) == 0)                                           		\
{   ALLOCN((PTR),TYPE,(NEW_N));}                            		\
    else									\
{								    		\
    (PTR) = (TYPE *)realloc((PTR),(NEW_N)*sizeof(TYPE));			\
    if (((PTR) == NULL) && ((NEW_N) != 0))					\
{									\
    fprintf(stderr, "Memory reallocation failed on line %d in %s\n", 	\
    __LINE__, __FILE__);                             		\
    fprintf(stderr, "  tried to reallocate %d->%d\n",       		\
    (OLD_N), (NEW_N));                              		\
    exit(-1);								\
}									\
    if ((NEW_N)>(OLD_N))							\
    memset((char *)(PTR)+(OLD_N)*sizeof(TYPE), 0,			\
    ((NEW_N)-(OLD_N))*sizeof(TYPE));				\
}										\
}

#define  ALLOCN(PTR,TYPE,N) 					\
{ (PTR) = (TYPE *) calloc(((unsigned)(N)),sizeof(TYPE));\
    if ((PTR) == NULL) {    				\
    fprintf(stderr, "Memory allocation failed on line %d in %s\n", \
    __LINE__, __FILE__);                           \
    exit(-1);                                             \
    }							\
}


#define FREE(PTR)  { free((PTR)); (PTR) = NULL; }
#endif


/*** declaration of routines ***/

PlyFile *ply_write(FILE *, int, const char **, int);
PlyFile *ply_open_for_writing(char *, int, const char **, int, float *);
void ply_describe_element(PlyFile *, char *, int, int, PlyProperty *);
void ply_describe_property(PlyFile *, char *, PlyProperty *);
void ply_element_count(PlyFile *, char *, int);
void ply_header_complete(PlyFile *);
void ply_put_element_setup(PlyFile *, const char *);
void ply_put_element(PlyFile *, void *);
void ply_put_comment(PlyFile *, char *);
void ply_put_obj_info(PlyFile *, char *);
PlyFile *ply_read(FILE *, int *, char ***);
PlyFile *ply_open_for_reading( char *, int *, char ***, int *, float *);
PlyProperty **ply_get_element_description(PlyFile *, char *, int*, int*);
void ply_get_element_setup( PlyFile *, char *, int, PlyProperty *);
int ply_get_property(PlyFile *, char *, PlyProperty *);
PlyOtherProp *ply_get_other_properties(PlyFile *, char *, int);
void ply_get_element(PlyFile *, void *);
char **ply_get_comments(PlyFile *, int *);
char **ply_get_obj_info(PlyFile *, int *);
void ply_close(PlyFile *);
void ply_get_info(PlyFile *, float *, int *);
PlyOtherElems *ply_get_other_element (PlyFile *, char *, int);
void ply_describe_other_elements ( PlyFile *, PlyOtherElems *);
void ply_put_other_elements (PlyFile *);
void ply_free_other_elements (PlyOtherElems *);
void ply_describe_other_properties(PlyFile *, PlyOtherProp *, int);

extern int equal_strings(const char *, const char *);

static PlyProperty vert_props[] = {
    {"x",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertexWithAlpha,x), 0, 0, 0, 0},
    {"y",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertexWithAlpha,y), 0, 0, 0, 0},
    {"z",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertexWithAlpha,z), 0, 0, 0, 0},
};

static PlyProperty colored_a_vert_props[] = {
    {"x",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertexWithAlpha,x), 0, 0, 0, 0},
    {"y",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertexWithAlpha,y), 0, 0, 0, 0},
    {"z",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertexWithAlpha,z), 0, 0, 0, 0},
    {"red",   PLY_UCHAR, PLY_UCHAR, offsetof(sPlyColoredVertexWithAlpha,red), 0, 0, 0, 0},
    {"green", PLY_UCHAR, PLY_UCHAR, offsetof(sPlyColoredVertexWithAlpha,green), 0, 0, 0, 0},
    {"blue",  PLY_UCHAR, PLY_UCHAR, offsetof(sPlyColoredVertexWithAlpha,blue), 0, 0, 0, 0},
    {"alpha", PLY_UCHAR, PLY_UCHAR, offsetof(sPlyColoredVertexWithAlpha,alpha), 0, 0, 0, 0}
};

static PlyProperty colored_vert_props[] = {
    {"x",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertex,x), 0, 0, 0, 0},
    {"y",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertex,y), 0, 0, 0, 0},
    {"z",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyColoredVertex,z), 0, 0, 0, 0},
    {"red",   PLY_UCHAR, PLY_UCHAR, offsetof(sPlyColoredVertex,red), 0, 0, 0, 0},
    {"green", PLY_UCHAR, PLY_UCHAR, offsetof(sPlyColoredVertex,green), 0, 0, 0, 0},
    {"blue",  PLY_UCHAR, PLY_UCHAR, offsetof(sPlyColoredVertex,blue), 0, 0, 0, 0},
};

static PlyProperty oriented_vert_props[] = {
    {"x",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedVertex,x ), 0, 0, 0, 0},
    {"y",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedVertex,y ), 0, 0, 0, 0},
    {"z",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedVertex,z ), 0, 0, 0, 0},
    {"nx", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedVertex,nx), 0, 0, 0, 0},
    {"ny", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedVertex,ny), 0, 0, 0, 0},
    {"nz", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedVertex,nz), 0, 0, 0, 0}
};

static PlyProperty oriented_colored_vert_props[] = {
    {"x",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredVertex,x ), 0, 0, 0, 0},
    {"y",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredVertex,y ), 0, 0, 0, 0},
    {"z",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredVertex,z ), 0, 0, 0, 0},
    {"nx", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredVertex,nx), 0, 0, 0, 0},
    {"ny", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredVertex,ny), 0, 0, 0, 0},
    {"nz", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredVertex,nz), 0, 0, 0, 0},
    {"red",   PLY_UCHAR, PLY_UCHAR, offsetof(sPlyOrientedColoredVertex,red), 0, 0, 0, 0},
    {"green", PLY_UCHAR, PLY_UCHAR, offsetof(sPlyOrientedColoredVertex,green), 0, 0, 0, 0},
    {"blue",  PLY_UCHAR, PLY_UCHAR, offsetof(sPlyOrientedColoredVertex,blue), 0, 0, 0, 0}
};

static PlyProperty oriented_colored_alpha_vert_props[] = {
    {"x",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredAlphaVertex,x ), 0, 0, 0, 0},
    {"y",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredAlphaVertex,y ), 0, 0, 0, 0},
    {"z",  PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredAlphaVertex,z ), 0, 0, 0, 0},
    {"nx", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredAlphaVertex,nx), 0, 0, 0, 0},
    {"ny", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredAlphaVertex,ny), 0, 0, 0, 0},
    {"nz", PLY_FLOAT, PLY_FLOAT, offsetof(sPlyOrientedColoredAlphaVertex,nz), 0, 0, 0, 0},
    {"red",   PLY_UCHAR, PLY_UCHAR, offsetof(sPlyOrientedColoredAlphaVertex,red), 0, 0, 0, 0},
    {"green", PLY_UCHAR, PLY_UCHAR, offsetof(sPlyOrientedColoredAlphaVertex,green), 0, 0, 0, 0},
    {"blue",  PLY_UCHAR, PLY_UCHAR, offsetof(sPlyOrientedColoredAlphaVertex,blue), 0, 0, 0, 0},
    {"alpha", PLY_UCHAR, PLY_UCHAR, offsetof(sPlyOrientedColoredAlphaVertex,alpha), 0, 0, 0, 0}
};

/*#ifndef WIN32
#ifdef __cplusplus
}
#endif
#endif*/
#endif // PLY_FILE_INCLUDED
